/*
 * linux/kernel/ptrace.c
 *
 * (C) Copyright 1999 Linus Torvalds
 *
 * Common interfaces for "ptrace()" which we do not want
 * to continually duplicate across every architecture.
 */

#include <linux/capability.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/pagemap.h>
#include <linux/smp_lock.h>
#include <linux/ptrace.h>
#include <linux/utrace.h>
#include <linux/security.h>
#include <linux/signal.h>
#include <linux/audit.h>
#include <linux/pid_namespace.h>
#include <linux/syscalls.h>

#include <asm/pgtable.h>
#include <asm/uaccess.h>


/*
 * Initialize a new task whose father had been ptraced.
 *
 * Called from copy_process().
 */
void ptrace_fork(struct task_struct *child, unsigned long clone_flags)
{
	arch_ptrace_fork(child, clone_flags);
}

/*
 * ptrace a task: make the debugger its new parent and
 * move it to the ptrace list.
 *
 * Must be called with the tasklist lock write-held.
 */
void __ptrace_link(struct task_struct *child, struct task_struct *new_parent)
{
	BUG_ON(!list_empty(&child->ptrace_entry));
	list_add(&child->ptrace_entry, &new_parent->ptraced);
	child->parent = new_parent;
}
 
#ifndef CONFIG_UTRACE_PTRACE
/*
 * Turn a tracing stop into a normal stop now, since with no tracer there
 * would be no way to wake it up with SIGCONT or SIGKILL.  If there was a
 * signal sent that would resume the child, but didn't because it was in
 * TASK_TRACED, resume it now.
 * Requires that irqs be disabled.
 */
static void ptrace_untrace(struct task_struct *child)
{
	spin_lock(&child->sighand->siglock);
	if (task_is_traced(child)) {
		if (child->signal->flags & SIGNAL_STOP_STOPPED) {
			__set_task_state(child, TASK_STOPPED);
		} else {
			signal_wake_up(child, 1);
		}
	}
	spin_unlock(&child->sighand->siglock);
}

static void ptrace_finish(struct task_struct *child)
{
	if (task_is_traced(child))
		ptrace_untrace(child);
}

static void ptrace_detach_task(struct task_struct *child, int sig)
{
	/* Architecture-specific hardware disable .. */
	ptrace_disable(child);
	clear_tsk_thread_flag(child, TIF_SYSCALL_TRACE);
}

static void utrace_engine_put(struct utrace_attached_engine *engine)
{
}

#else  /* CONFIG_UTRACE_PTRACE */

static const struct utrace_engine_ops ptrace_utrace_ops; /* forward decl */

static void ptrace_detach_task(struct task_struct *child, int sig)
{
	enum utrace_resume_action action = UTRACE_DETACH;
	struct utrace_attached_engine *engine;
	int ret;

	engine = utrace_attach_task(child, UTRACE_ATTACH_MATCH_OPS,
				    &ptrace_utrace_ops, NULL);
	if (unlikely(IS_ERR(engine)))
		return;

	/*
	 * When leaving a parting signal to deliver, we cannot detach our
	 * utrace engine yet.  We need it there to get a report_signal()
	 * callback that can inject our parting signal.  Instead, we'll
	 * mark our engine by clearing its data pointer.  Our callbacks
	 * will see this and know it's a PTRACE_DETACH still in progress.
	 */
	if (sig) {
		engine->data = NULL;
		action = UTRACE_INTERRUPT;
	}

	ret = utrace_control(child, engine, action);
	WARN_ON(ret && ret != -EINPROGRESS &&
		ret != -ESRCH && ret != -EALREADY);

	utrace_engine_put(engine);
}

void ptrace_utrace_exit(struct task_struct *task)
{
	struct task_struct *child;
	read_lock(&tasklist_lock);
	list_for_each_entry(child, &task->ptraced, ptrace_entry)
		ptrace_detach_task(child, 0);
	read_unlock(&tasklist_lock);
}

static void ptrace_finish(struct task_struct *child)
{
}

#endif	/* !CONFIG_UTRACE_PTRACE */

/*
 * unptrace a task: move it back to its original parent and
 * remove it from the ptrace list.
 *
 * Must be called with the tasklist lock write-held.
 */
void __ptrace_unlink(struct task_struct *child)
{
	BUG_ON(!child->ptrace);

	child->ptrace = 0;
	child->parent = child->real_parent;
	list_del_init(&child->ptrace_entry);

	arch_ptrace_untrace(child);
	ptrace_finish(child);
}

#ifndef CONFIG_UTRACE_PTRACE

/*
 * Check that we have indeed attached to the thing..
 */
int ptrace_check_attach(struct task_struct *child, int kill)
{
	int ret = -ESRCH;

	/*
	 * We take the read lock around doing both checks to close a
	 * possible race where someone else was tracing our child and
	 * detached between these two checks.  After this locked check,
	 * we are sure that this is our traced child and that can only
	 * be changed by us so it's not changing right after this.
	 */
	read_lock(&tasklist_lock);
	if ((child->ptrace & PT_PTRACED) && child->parent == current) {
		ret = 0;
		/*
		 * child->sighand can't be NULL, release_task()
		 * does ptrace_unlink() before __exit_signal().
		 */
		spin_lock_irq(&child->sighand->siglock);
		if (task_is_stopped(child))
			child->state = TASK_TRACED;
		else if (!task_is_traced(child) && !kill)
			ret = -ESRCH;
		spin_unlock_irq(&child->sighand->siglock);
	}
	read_unlock(&tasklist_lock);

	if (!ret && !kill)
		ret = wait_task_inactive(child, TASK_TRACED) ? 0 : -ESRCH;

	/* All systems go.. */
	return ret;
}

static struct utrace_attached_engine *prepare_ptrace_attach(
	struct task_struct *child, struct task_struct *parent)
{
	return NULL;
}

static int finish_ptrace_attach(struct task_struct *task,
				 struct utrace_attached_engine *engine,
				 int retval)
{
	return retval;
}

static int ptrace_update_utrace(struct task_struct *task,
				struct utrace_attached_engine *engine)
{
	return 0;
}

#else	/* CONFIG_UTRACE_PTRACE */

static int ptrace_update_utrace(struct task_struct *task,
				struct utrace_attached_engine *engine)
{
	unsigned long events;

	/*
	 * We need this for resume handling.
	 */
	events = UTRACE_EVENT(QUIESCE);

	/*
	 * These events are always reported.
	 */
	events |= UTRACE_EVENT(EXEC) | UTRACE_EVENT_SIGNAL_ALL;

	/*
	 * We always have to examine clone events to check for CLONE_PTRACE.
	 */
	events |= UTRACE_EVENT(CLONE);

	/*
	 * PTRACE_SETOPTIONS can request more events.
	 */
	if (task->ptrace & PT_TRACE_EXIT)
		events |= UTRACE_EVENT(EXIT);

	if (!engine) {
		int ret;
		engine = utrace_attach_task(task, UTRACE_ATTACH_MATCH_OPS,
					    &ptrace_utrace_ops, NULL);
		if (IS_ERR(engine))
			return -ESRCH;
		ret = utrace_set_events(task, engine, events);
		utrace_engine_put(engine);
		return ret;
	}

	return utrace_set_events(task, engine, events);
}

static int ptrace_unsafe_exec(struct utrace_attached_engine *engine,
			      struct task_struct *task)
{
	if (task->ptrace & PT_PTRACE_CAP)
		return LSM_UNSAFE_PTRACE_CAP;

	return LSM_UNSAFE_PTRACE;
}

static struct task_struct *ptrace_tracer_task(
	struct utrace_attached_engine *engine, struct task_struct *target)
{
	return target->parent;
}

static void ptrace_set_action(struct task_struct *task,
			      enum utrace_resume_action action,
			      enum utrace_syscall_action syscall)
{
	task->ptrace &= ~((UTRACE_SYSCALL_MASK | UTRACE_RESUME_MASK) << 16);
	task->ptrace |= ((UTRACE_RESUME - action) | syscall) << 16;
}

static enum utrace_resume_action ptrace_resume_action(struct task_struct *task)
{
	return UTRACE_RESUME - ((task->ptrace >> 16) & UTRACE_RESUME_MASK);
}

static enum utrace_syscall_action ptrace_syscall_action(
	struct task_struct *task)
{
	return (task->ptrace >> 16) & UTRACE_SYSCALL_MASK;
}

/*
 * Remember which event stop this was.  We have to keep a record that
 * won't be wiped by ptrace_do_wait() when @task->exit_code is cleared.
 * So we store the PTRACE_EVENT_* value in a nybble of @task->ptrace.
 */
static void ptrace_set_stop_event(struct task_struct *task, int event)
{
	task->ptrace &= 0xf0ffffffU;
	task->ptrace |= event << 24;
}

static int ptrace_stop_event(struct task_struct *task)
{
	return (task->ptrace >> 24) & 0xf;
}

/*
 * This doesn't really exist in the user API, but we pick an unused value
 * encoded with ptrace_set_stop_event() to record when the last stop was
 * for a syscall stop.
 */
#define PTRACE_EVENT_SYSCALL	0xf

static u32 utrace_ptrace_report(u32 action, struct task_struct *task,
				int event, int code)
{
	/*
	 * Special kludge magic in utrace.c (utrace_stop) sees this
	 * and calls do_notify_parent_cldstop() for us.  This kludge
	 * is necessary to keep that wakeup after we enter TASK_TRACED.
	 */
	ptrace_set_action(task, UTRACE_STOP, 0);

	/*
	 * If we already have a pending stop event, then don't override it.
	 * We'll simulate this stop when the pending one is cleared,
	 * in ptrace_resume(), below.
	 */
	if (ptrace_stop_event(task) == 0) {
		ptrace_set_stop_event(task, event);
		task->exit_code = code;
	}

	return action | UTRACE_STOP;
}

static u32 utrace_ptrace_event(struct task_struct *task,
			       int event, unsigned long msg)
{
	task->ptrace_message = msg;
	return utrace_ptrace_report(0, task, event, (event << 8) | SIGTRAP);
}

static u32 ptrace_report_exec(enum utrace_resume_action action,
			      struct utrace_attached_engine *engine,
			      struct task_struct *task,
			      const struct linux_binfmt *fmt,
			      const struct linux_binprm *bprm,
			      struct pt_regs *regs)
{
	if (task->ptrace & PT_TRACE_EXEC)
		return utrace_ptrace_event(task, PTRACE_EVENT_EXEC, 0);

	/*
	 * Old-fashioned ptrace'd exec just posts a plain signal.
	 */
	send_sig(SIGTRAP, task, 0);
	return UTRACE_RESUME;
}

static u32 ptrace_report_exit(enum utrace_resume_action action,
			      struct utrace_attached_engine *engine,
			      struct task_struct *task,
			      long orig_code, long *code)
{
	return utrace_ptrace_event(task, PTRACE_EVENT_EXIT, *code);
}

static u32 ptrace_report_clone(enum utrace_resume_action action,
			       struct utrace_attached_engine *engine,
			       struct task_struct *parent,
			       unsigned long clone_flags,
			       struct task_struct *child)
{
	int event;
	struct utrace_attached_engine *child_engine;

	if ((clone_flags & CLONE_VFORK) &&
	    (parent->ptrace & PT_TRACE_VFORK_DONE)) {
		/*
		 * Stash the child PID for a PTRACE_EVENT_VFORK_DONE report,
		 * even if we don't report this clone event itself.  Mark
		 * that we vfork'd even if we don't actually stop here, so
		 * that we'll know to do the vfork-done report.  (This
		 * matters when is PTRACE_O_TRACEVFORKDONE set but
		 * PTRACE_O_TRACEVFORK is not set.)
		 */
		int rc = utrace_control(parent, engine, UTRACE_REPORT);
		WARN_ON(rc);
		parent->ptrace_message = child->pid;
		ptrace_set_stop_event(parent, PTRACE_EVENT_VFORK);
	}

	if (clone_flags & CLONE_UNTRACED)
		return UTRACE_RESUME;

	event = 0;
	if (clone_flags & CLONE_VFORK) {
		if (parent->ptrace & PT_TRACE_VFORK)
			event = PTRACE_EVENT_VFORK;
	} else if ((clone_flags & CSIGNAL) != SIGCHLD) {
		if (parent->ptrace & PT_TRACE_CLONE)
			event = PTRACE_EVENT_CLONE;
	} else if (parent->ptrace & PT_TRACE_FORK) {
		event = PTRACE_EVENT_FORK;
	}

	/*
	 * Any of these reports implies auto-attaching the new child.
	 * So does CLONE_PTRACE, even with no event to report.
	 */
	if (event || (clone_flags & CLONE_PTRACE)) {
		child_engine = utrace_attach_task(child, UTRACE_ATTACH_CREATE |
						  UTRACE_ATTACH_EXCLUSIVE |
						  UTRACE_ATTACH_MATCH_OPS,
						  &ptrace_utrace_ops,
						  parent->parent);
		if (unlikely(IS_ERR(child_engine))) {
			WARN_ON(1);	/* XXX */
		} else {
			BUG_ON(!child->ptrace);
			/* XXX already set by old ptrace code
			   task_lock(child);
			   child->ptrace = parent->ptrace;
			   child->parent = parent->parent;
			   task_unlock(child);
			*/
			ptrace_update_utrace(child, child_engine);
			utrace_engine_put(child_engine);
		}
	}

	if (event)
		return utrace_ptrace_event(parent, event, child->pid);

	return UTRACE_RESUME;
}

static u32 ptrace_report_syscall(u32 action, struct task_struct *task)
{
	int code = SIGTRAP;
	if (task->ptrace & PT_TRACESYSGOOD)
		code |= 0x80;
	return utrace_ptrace_report(action, task, PTRACE_EVENT_SYSCALL, code);
}

static u32 ptrace_report_syscall_entry(u32 action,
				       struct utrace_attached_engine *engine,
				       struct task_struct *task,
				       struct pt_regs *regs)
{
	/*
	 * If we're doing PTRACE_SYSEMU, just punt here and report
	 * at the exit stop instead.
	 */
	if (ptrace_syscall_action(task))
		return UTRACE_SYSCALL_ABORT | UTRACE_RESUME;

	return ptrace_report_syscall(UTRACE_SYSCALL_RUN, task);
}

/*
 * Report a vfork-done event to precede another event report.
 * There is no utrace callback associated directly with vfork-done.
 * Instead, after seeing CLONE_VFORK in report_clone() we notice
 * the following syscall-exit, signal, or quiesce callback and
 * inject the PTRACE_EVENT_VFORK_DONE stop there.
 */
static u32 ptrace_report_vfork_done(struct task_struct *task)
{
	/*
	 * We got here after a vfork, when vfork-done tracing
	 * was enabled.
	 */
	ptrace_set_stop_event(task, 0);
	return utrace_ptrace_event(task, PTRACE_EVENT_VFORK_DONE,
				   task->ptrace_message);
}

static u32 ptrace_report_syscall_exit(enum utrace_resume_action action,
				      struct utrace_attached_engine *engine,
				      struct task_struct *task,
				      struct pt_regs *regs)
{
	if (unlikely(ptrace_stop_event(task) == PTRACE_EVENT_VFORK))
		return ptrace_report_vfork_done(task);

	if (unlikely(ptrace_syscall_action(task)) &&
	    unlikely(ptrace_resume_action(task) == UTRACE_SINGLESTEP))
		/*
		 * This is PTRACE_SYSEMU_SINGLESTEP.
		 * Kludge: Prevent arch code from sending a SIGTRAP
		 * after tracehook_report_syscall_exit() returns.
		 */
		user_disable_single_step(task);

	return ptrace_report_syscall(0, task);
}

static u32 ptrace_resumed(struct task_struct *task,
			  struct utrace_attached_engine *engine,
			  siginfo_t *info, struct k_sigaction *return_ka)
{
	enum utrace_resume_action resume = ptrace_resume_action(task);

	if (unlikely(ptrace_stop_event(task) == PTRACE_EVENT_VFORK))
		return UTRACE_SIGNAL_REPORT | ptrace_report_vfork_done(task);

	/*
	 * If PTRACE_DETACH had a signal for us to deliver,
	 * it didn't detach the engine, but marked it for us.
	 */
	if (!engine->data) {
		read_lock(&tasklist_lock);
		if (task->ptrace & PT_PTRACED)
			engine->data = task->parent;
		else
			resume = UTRACE_DETACH;
		read_unlock(&tasklist_lock);
	}

	/*
	 * If we're stopping, or we haven't reported any signal,
	 * then we're all done for now.
	 */
	if (resume == UTRACE_STOP || !task->last_siginfo)
		return UTRACE_SIGNAL_REPORT | resume;

	/*
	 * We're resuming.  If there's no signal to deliver, just go.
	 * If we were given a signal, deliver it now.
	 */
	WARN_ON(task->last_siginfo != info);
	task->last_siginfo = NULL;
	if (!task->exit_code)
		return UTRACE_SIGNAL_REPORT | resume;

	/* Update the siginfo structure if the signal has
	   changed.  If the debugger wanted something
	   specific in the siginfo structure then it should
	   have updated *info via PTRACE_SETSIGINFO.  */
	if (task->exit_code != info->si_signo) {
		info->si_signo = task->exit_code;
		info->si_errno = 0;
		info->si_code = SI_USER;
		info->si_pid = task_pid_vnr(task->parent);
		info->si_uid = __task_cred(task->parent)->uid;
	}

	task->exit_code = 0;

	spin_lock_irq(&task->sighand->siglock);
	*return_ka = task->sighand->action[info->si_signo - 1];
	spin_unlock_irq(&task->sighand->siglock);

	return UTRACE_SIGNAL_DELIVER | resume;
}

static u32 ptrace_report_signal(u32 action,
				struct utrace_attached_engine *engine,
				struct task_struct *task,
				struct pt_regs *regs,
				siginfo_t *info,
				const struct k_sigaction *orig_ka,
				struct k_sigaction *return_ka)
{
	switch (utrace_signal_action(action)) {
	default:
		WARN_ON(ptrace_stop_event(task) && info->si_signo != SIGKILL);
		break;
	case UTRACE_SIGNAL_HANDLER:
		WARN_ON(ptrace_stop_event(task));
		/*
		 * A handler was set up.  If we are stepping, pretend
		 * another SIGTRAP arrived.
		 */
		if (ptrace_resume_action(task) == UTRACE_SINGLESTEP ||
		    ptrace_resume_action(task) == UTRACE_BLOCKSTEP) {
			memset(info, 0, sizeof *info);
			info->si_signo = SIGTRAP;
			info->si_code = SIGTRAP;
			info->si_pid = task_pid_vnr(task);
			info->si_uid = __task_cred(task->parent)->uid;
			break;
		}
		/* Fall through.  */
	case UTRACE_SIGNAL_REPORT:
		/*
		 * This is not a new signal, but just a notification we
		 * asked for.  Either we're stopping after another report
		 * like exec or syscall, or we're resuming.
		 */
		return ptrace_resumed(task, engine, info, return_ka);
	}

	WARN_ON(!valid_signal(info->si_signo));
	task->last_siginfo = info;
	return utrace_ptrace_report(UTRACE_SIGNAL_IGN, task, 0, info->si_signo);
}

static u32 ptrace_report_quiesce(u32 action,
				 struct utrace_attached_engine *engine,
				 struct task_struct *task,
				 unsigned long event)
{
	if (event == 0) {
		task->last_siginfo = NULL;

		if (unlikely(ptrace_stop_event(task) == PTRACE_EVENT_VFORK))
			return ptrace_report_vfork_done(task);
	}

	return ptrace_resume_action(task);
}

static const struct utrace_engine_ops ptrace_utrace_ops = {
	.tracer_task = ptrace_tracer_task,
	.unsafe_exec = ptrace_unsafe_exec,
	.report_signal = ptrace_report_signal,
	.report_quiesce = ptrace_report_quiesce,
	.report_exec = ptrace_report_exec,
	.report_exit = ptrace_report_exit,
	.report_clone = ptrace_report_clone,
	.report_syscall_entry = ptrace_report_syscall_entry,
	.report_syscall_exit = ptrace_report_syscall_exit,
};

/*
 * Detach the utrace engine on error.  On success, set @engine->data to
 * the ptracer's task_struct pointer.  This distinguishes an engine for
 * attached ptrace from an engine left behind after a PTRACE_DETACH call
 * that left a parting signal to deliver.
 */
static int finish_ptrace_attach(struct task_struct *task,
				struct utrace_attached_engine *engine,
				int retval)
{
	if (retval) {
		int error = utrace_control(task, engine, UTRACE_DETACH);
		WARN_ON(error && error != -ESRCH && error != -EALREADY);
	} else {
		engine->data = task->parent;
	}
	utrace_engine_put(engine);
	return retval;
}

/*
 * Attach a utrace engine for ptrace and set up its event mask.
 * Returns the engine pointer or an IS_ERR() pointer.
 */
static struct utrace_attached_engine *prepare_ptrace_attach(
	struct task_struct *child, struct task_struct *parent)
{
	struct utrace_attached_engine *engine;
restart:
	engine = utrace_attach_task(child, UTRACE_ATTACH_CREATE |
				    UTRACE_ATTACH_EXCLUSIVE |
				    UTRACE_ATTACH_MATCH_OPS,
				    &ptrace_utrace_ops, parent);
	if (engine == ERR_PTR(-EEXIST)) {
		/*
		 * There is already some ptrace engine attached.  Either
		 * this attach will fail because another ptracer is
		 * already attached, or else this engine was left behind
		 * by PTRACE_DETACH to deliver a signal.
		 */
		engine = utrace_attach_task(child,
					    UTRACE_ATTACH_MATCH_OPS,
					    &ptrace_utrace_ops, NULL);
		if (IS_ERR(engine)) {
			BUG_ON(PTR_ERR(engine) != -ENOENT);
			goto restart;
		}
		if (engine->data) {
			utrace_engine_put(engine);
			engine = ERR_PTR(-EEXIST);
		}
	}
	if (IS_ERR(engine)) {
		if (engine != ERR_PTR(-ESRCH))
			engine = ERR_PTR(-EPERM);
	} else {
		int ret = ptrace_update_utrace(child, engine);
		if (ret == -EINPROGRESS) {
			ret = utrace_barrier(child, engine);
			utrace_engine_put(engine);
			if (ret != -ERESTARTSYS)
				goto restart;
			return ERR_PTR(-ERESTARTNOINTR);
		}
		if (ret)
			engine = ERR_PTR(finish_ptrace_attach(child, engine,
							      -ESRCH));
	}
	return engine;
}

int ptrace_check_attach(struct task_struct *child, int kill)
{
	struct utrace_attached_engine *engine;
	struct utrace_examiner exam;
	int ret;

	engine = utrace_attach_task(child, UTRACE_ATTACH_MATCH_OPS,
				    &ptrace_utrace_ops, NULL);
	if (IS_ERR(engine))
		return -ESRCH;

	/*
	 * Make sure our engine has already stopped the child.
	 * Then wait for it to be off the CPU.
	 */
	ret = 0;
	if (utrace_control(child, engine, UTRACE_STOP) ||
	    utrace_prepare_examine(child, engine, &exam))
		ret = -ESRCH;

	utrace_engine_put(engine);

	return ret;
}

#endif	/* !CONFIG_UTRACE_PTRACE */

int __ptrace_may_access(struct task_struct *task, unsigned int mode)
{
	const struct cred *cred = current_cred(), *tcred;

	/* May we inspect the given task?
	 * This check is used both for attaching with ptrace
	 * and for allowing access to sensitive information in /proc.
	 *
	 * ptrace_attach denies several cases that /proc allows
	 * because setting up the necessary parent/child relationship
	 * or halting the specified task is impossible.
	 */
	int dumpable = 0;
	/* Don't let security modules deny introspection */
	if (task == current)
		return 0;
	rcu_read_lock();
	tcred = __task_cred(task);
	if ((cred->uid != tcred->euid ||
	     cred->uid != tcred->suid ||
	     cred->uid != tcred->uid  ||
	     cred->gid != tcred->egid ||
	     cred->gid != tcred->sgid ||
	     cred->gid != tcred->gid) &&
	    !capable(CAP_SYS_PTRACE)) {
		rcu_read_unlock();
		return -EPERM;
	}
	rcu_read_unlock();
	smp_rmb();
	if (task->mm)
		dumpable = get_dumpable(task->mm);
	if (!dumpable && !capable(CAP_SYS_PTRACE))
		return -EPERM;

	return security_ptrace_may_access(task, mode);
}

bool ptrace_may_access(struct task_struct *task, unsigned int mode)
{
	int err;
	task_lock(task);
	err = __ptrace_may_access(task, mode);
	task_unlock(task);
	return (!err ? true : false);
}

int ptrace_attach(struct task_struct *task)
{
	int retval;
	unsigned long flags;
	struct utrace_attached_engine *engine;

	audit_ptrace(task);

	retval = -EPERM;
	if (same_thread_group(task, current))
		return retval;

	engine = prepare_ptrace_attach(task, current);
	if (unlikely(IS_ERR(engine)))
		return PTR_ERR(engine);

	/* Protect exec's credential calculations against our interference;
	 * SUID, SGID and LSM creds get determined differently under ptrace.
	 */
	retval = mutex_lock_interruptible(&current->cred_exec_mutex);
	if (retval  < 0)
		goto out;

	retval = -EPERM;
repeat:
	/*
	 * Nasty, nasty.
	 *
	 * We want to hold both the task-lock and the
	 * tasklist_lock for writing at the same time.
	 * But that's against the rules (tasklist_lock
	 * is taken for reading by interrupts on other
	 * cpu's that may have task_lock).
	 */
	task_lock(task);
	if (!write_trylock_irqsave(&tasklist_lock, flags)) {
		task_unlock(task);
		do {
			cpu_relax();
		} while (!write_can_lock(&tasklist_lock));
		goto repeat;
	}

	if (!task->mm)
		goto bad;
	/* the same process cannot be attached many times */
	if (task->ptrace & PT_PTRACED)
		goto bad;
	retval = __ptrace_may_access(task, PTRACE_MODE_ATTACH);
	if (retval)
		goto bad;

	/* Go */
	task->ptrace |= PT_PTRACED;
	if (capable(CAP_SYS_PTRACE))
		task->ptrace |= PT_PTRACE_CAP;

	__ptrace_link(task, current);

	send_sig_info(SIGSTOP, SEND_SIG_FORCED, task);
bad:
	write_unlock_irqrestore(&tasklist_lock, flags);
	task_unlock(task);
	mutex_unlock(&current->cred_exec_mutex);
out:
	return finish_ptrace_attach(task, engine, retval);
}

static inline void __ptrace_detach(struct task_struct *child, unsigned int data)
{
	child->exit_code = data;
	/* .. re-parent .. */
	__ptrace_unlink(child);
	/* .. and wake it up. */
	if (child->exit_state != EXIT_ZOMBIE)
		wake_up_process(child);
}

int ptrace_detach(struct task_struct *child, unsigned int data)
{
	if (!valid_signal(data))
		return -EIO;

	ptrace_detach_task(child, data);

	write_lock_irq(&tasklist_lock);
	/* protect against de_thread()->release_task() */
	if (child->ptrace)
		__ptrace_detach(child, data);
	write_unlock_irq(&tasklist_lock);

	return 0;
}

int ptrace_readdata(struct task_struct *tsk, unsigned long src, char __user *dst, int len)
{
	int copied = 0;

	while (len > 0) {
		char buf[128];
		int this_len, retval;

		this_len = (len > sizeof(buf)) ? sizeof(buf) : len;
		retval = access_process_vm(tsk, src, buf, this_len, 0);
		if (!retval) {
			if (copied)
				break;
			return -EIO;
		}
		if (copy_to_user(dst, buf, retval))
			return -EFAULT;
		copied += retval;
		src += retval;
		dst += retval;
		len -= retval;			
	}
	return copied;
}

int ptrace_writedata(struct task_struct *tsk, char __user *src, unsigned long dst, int len)
{
	int copied = 0;

	while (len > 0) {
		char buf[128];
		int this_len, retval;

		this_len = (len > sizeof(buf)) ? sizeof(buf) : len;
		if (copy_from_user(buf, src, this_len))
			return -EFAULT;
		retval = access_process_vm(tsk, dst, buf, this_len, 1);
		if (!retval) {
			if (copied)
				break;
			return -EIO;
		}
		copied += retval;
		src += retval;
		dst += retval;
		len -= retval;			
	}
	return copied;
}

static int ptrace_setoptions(struct task_struct *child, long data)
{
	child->ptrace &= ~PT_TRACE_MASK;

	if (data & PTRACE_O_TRACESYSGOOD)
		child->ptrace |= PT_TRACESYSGOOD;

	if (data & PTRACE_O_TRACEFORK)
		child->ptrace |= PT_TRACE_FORK;

	if (data & PTRACE_O_TRACEVFORK)
		child->ptrace |= PT_TRACE_VFORK;

	if (data & PTRACE_O_TRACECLONE)
		child->ptrace |= PT_TRACE_CLONE;

	if (data & PTRACE_O_TRACEEXEC)
		child->ptrace |= PT_TRACE_EXEC;

	if (data & PTRACE_O_TRACEVFORKDONE)
		child->ptrace |= PT_TRACE_VFORK_DONE;

	if (data & PTRACE_O_TRACEEXIT)
		child->ptrace |= PT_TRACE_EXIT;

	ptrace_update_utrace(child, NULL);

	return (data & ~PTRACE_O_MASK) ? -EINVAL : 0;
}

static int ptrace_getsiginfo(struct task_struct *child, siginfo_t *info)
{
	int error = -ESRCH;

	read_lock(&tasklist_lock);
	if (likely(child->sighand != NULL)) {
		error = -EINVAL;
		spin_lock_irq(&child->sighand->siglock);
		if (likely(child->last_siginfo != NULL)) {
			*info = *child->last_siginfo;
			error = 0;
		}
		spin_unlock_irq(&child->sighand->siglock);
	}
	read_unlock(&tasklist_lock);
	return error;
}

static int ptrace_setsiginfo(struct task_struct *child, const siginfo_t *info)
{
	int error = -ESRCH;

	read_lock(&tasklist_lock);
	if (likely(child->sighand != NULL)) {
		error = -EINVAL;
		spin_lock_irq(&child->sighand->siglock);
		if (likely(child->last_siginfo != NULL)) {
			*child->last_siginfo = *info;
			error = 0;
		}
		spin_unlock_irq(&child->sighand->siglock);
	}
	read_unlock(&tasklist_lock);
	return error;
}


#ifdef PTRACE_SINGLESTEP
#define is_singlestep(request)		((request) == PTRACE_SINGLESTEP)
#else
#define is_singlestep(request)		0
#endif

#ifdef PTRACE_SINGLEBLOCK
#define is_singleblock(request)		((request) == PTRACE_SINGLEBLOCK)
#else
#define is_singleblock(request)		0
#endif

#ifdef PTRACE_SYSEMU
#define is_sysemu_singlestep(request)	((request) == PTRACE_SYSEMU_SINGLESTEP)
#else
#define is_sysemu_singlestep(request)	0
#endif

#ifndef CONFIG_UTRACE_PTRACE
static int ptrace_resume(struct task_struct *child, long request, long data)
{
	if (!valid_signal(data))
		return -EIO;

	if (request == PTRACE_SYSCALL)
		set_tsk_thread_flag(child, TIF_SYSCALL_TRACE);
	else
		clear_tsk_thread_flag(child, TIF_SYSCALL_TRACE);

#ifdef TIF_SYSCALL_EMU
	if (request == PTRACE_SYSEMU || request == PTRACE_SYSEMU_SINGLESTEP)
		set_tsk_thread_flag(child, TIF_SYSCALL_EMU);
	else
		clear_tsk_thread_flag(child, TIF_SYSCALL_EMU);
#endif

	if (is_singleblock(request)) {
		if (unlikely(!arch_has_block_step()))
			return -EIO;
		user_enable_block_step(child);
	} else if (is_singlestep(request) || is_sysemu_singlestep(request)) {
		if (unlikely(!arch_has_single_step()))
			return -EIO;
		user_enable_single_step(child);
	}
	else
		user_disable_single_step(child);

	child->exit_code = data;
	wake_up_process(child);

	return 0;
}
#else  /* CONFIG_UTRACE_PTRACE */
static int ptrace_resume(struct task_struct *child, long request, long data)
{
	struct utrace_attached_engine *engine;
	enum utrace_resume_action action;
	enum utrace_syscall_action syscall;
	int ret = 0;

	if (!valid_signal(data))
		return -EIO;

	engine = utrace_attach_task(child, UTRACE_ATTACH_MATCH_OPS,
				    &ptrace_utrace_ops, NULL);
	if (IS_ERR(engine))
		return -ESRCH;

	syscall = UTRACE_SYSCALL_RUN;
#ifdef PTRACE_SYSEMU
	if (request == PTRACE_SYSEMU || request == PTRACE_SYSEMU_SINGLESTEP)
		syscall = UTRACE_SYSCALL_ABORT;
#endif

	if (syscall != UTRACE_SYSCALL_RUN || request == PTRACE_SYSCALL) {
		if (!(engine->flags & UTRACE_EVENT_SYSCALL) &&
		    utrace_set_events(child, engine,
				      engine->flags | UTRACE_EVENT_SYSCALL))
			ret = -ESRCH;
	} else if (engine->flags & UTRACE_EVENT(SYSCALL_ENTRY)) {
		if (utrace_set_events(child, engine,
				      engine->flags & ~UTRACE_EVENT_SYSCALL))
			ret = -ESRCH;
	}

	action = UTRACE_RESUME;
	if (is_singleblock(request)) {
		if (unlikely(!arch_has_block_step()))
			ret = -EIO;
		action = UTRACE_BLOCKSTEP;
	} else if (is_singlestep(request) || is_sysemu_singlestep(request)) {
		if (unlikely(!arch_has_single_step()))
			ret = -EIO;
		action = UTRACE_SINGLESTEP;
	}

	if (!ret) {
		/*
		 * If there was a non-signal event reported last time,
		 * we may need to simulate another event that follows
		 * it, or do some other special case.  We'll set @event
		 * nonzero here if there is an event to simulate.
		 */
		int event = ptrace_stop_event(child);
		ptrace_set_stop_event(child, 0);

		switch (event) {
		case 0:
			break;

		default:
			WARN_ON(1);
		case PTRACE_EVENT_EXIT:
			event = 0;
			break;

		case PTRACE_EVENT_FORK:
		case PTRACE_EVENT_CLONE:
		case PTRACE_EVENT_EXEC:
		case PTRACE_EVENT_VFORK_DONE:
			/*
			 * At these events, @data is ignored.
			 * After these there might be a syscall-exit stop.
			 */
			if (request == PTRACE_SYSCALL) {
				data = SIGTRAP;
				if (child->ptrace & PT_TRACESYSGOOD)
					data |= 0x80;
				ptrace_set_stop_event(child,
						      PTRACE_EVENT_SYSCALL);
			} else {
				data = 0;
				event = 0;
			}
			break;

		case PTRACE_EVENT_SYSCALL:
			/*
			 * After a syscall stop, @data is treated specially.
			 * It just queues the signal.
			 */
			if (data)
				send_sig(data, child, 1);
			data = 0;
			event = 0;
			break;

		case PTRACE_EVENT_VFORK:
			/*
			 * For this event, @data is ignored.  We did have a
			 * utrace stop right after, so on resuming there
			 * can be a normal syscall-exit stop next if needed.
			 * But we may need to simulate a vfork-done stop,
			 * so we leave the marker in place.
			 */
			data = 0;
			event = 0;
			if (child->ptrace & PT_TRACE_VFORK_DONE) {
				ptrace_set_stop_event(child,
						      PTRACE_EVENT_VFORK);
				action = UTRACE_REPORT;
			}
			break;
		}

		child->exit_code = data;

		ptrace_set_action(child, action, syscall);

		/*
		 * Whatever action we want to resume with, we need to make
		 * sure the child gets into ptrace_report_quiesce() so that
		 * it clears last_siginfo before going back to user mode.
		 */
		if (child->last_siginfo)
			action = UTRACE_REPORT;

		if (event) {
			/*
			 * We have a stacked event.  That is, @child is
			 * already at the last utrace stop point before
			 * returning to user mode.  But the ptrace API
			 * wants multiple stops on the way out.  Here the
			 * original ptrace would resume the child and have
			 * to stop again almost immediately.  Instead we
			 * just leave it stopped and simulate the wake-up
			 * and new stop by notifying the parent, i.e.
			 * current's own thread group.
			 */
			read_lock(&tasklist_lock);
			do_notify_parent_cldstop(child, CLD_TRAPPED);
			read_unlock(&tasklist_lock);
		} else {
			if (task_is_stopped(child)) {
				spin_lock_irq(&child->sighand->siglock);
				child->signal->flags &= ~SIGNAL_STOP_STOPPED;
				spin_unlock_irq(&child->sighand->siglock);
			}

			/*
			 * To resume with a signal we must arrange
			 * to enter ptrace_report_signal().
			 */
			if (data)
				action = UTRACE_INTERRUPT;

			if (utrace_control(child, engine, action))
				ret = -ESRCH;
		}
	}

	utrace_engine_put(engine);

	return ret;
}
#endif	/* !CONFIG_UTRACE_PTRACE */

int ptrace_request(struct task_struct *child, long request,
		   long addr, long data)
{
	int ret = -EIO;
	siginfo_t siginfo;

	switch (request) {
	case PTRACE_PEEKTEXT:
	case PTRACE_PEEKDATA:
		return generic_ptrace_peekdata(child, addr, data);
	case PTRACE_POKETEXT:
	case PTRACE_POKEDATA:
		return generic_ptrace_pokedata(child, addr, data);

#ifdef PTRACE_OLDSETOPTIONS
	case PTRACE_OLDSETOPTIONS:
#endif
	case PTRACE_SETOPTIONS:
		ret = ptrace_setoptions(child, data);
		break;
	case PTRACE_GETEVENTMSG:
		ret = put_user(child->ptrace_message, (unsigned long __user *) data);
		break;

	case PTRACE_GETSIGINFO:
		ret = ptrace_getsiginfo(child, &siginfo);
		if (!ret)
			ret = copy_siginfo_to_user((siginfo_t __user *) data,
						   &siginfo);
		break;

	case PTRACE_SETSIGINFO:
		if (copy_from_user(&siginfo, (siginfo_t __user *) data,
				   sizeof siginfo))
			ret = -EFAULT;
		else
			ret = ptrace_setsiginfo(child, &siginfo);
		break;

	case PTRACE_DETACH:	 /* detach a process that was attached. */
		ret = ptrace_detach(child, data);
		break;

#ifdef PTRACE_SINGLESTEP
	case PTRACE_SINGLESTEP:
#endif
#ifdef PTRACE_SINGLEBLOCK
	case PTRACE_SINGLEBLOCK:
#endif
#ifdef PTRACE_SYSEMU
	case PTRACE_SYSEMU:
	case PTRACE_SYSEMU_SINGLESTEP:
#endif
	case PTRACE_SYSCALL:
	case PTRACE_CONT:
		return ptrace_resume(child, request, data);

	case PTRACE_KILL:
		if (child->exit_state)	/* already dead */
			return 0;
		return ptrace_resume(child, request, SIGKILL);

	default:
		break;
	}

	return ret;
}

/**
 * ptrace_traceme  --  helper for PTRACE_TRACEME
 *
 * Performs checks and sets PT_PTRACED.
 * Should be used by all ptrace implementations for PTRACE_TRACEME.
 */
int ptrace_traceme(void)
{
	int ret = -EPERM;
	struct utrace_attached_engine *engine;

	engine = prepare_ptrace_attach(current, current->parent);
	if (unlikely(IS_ERR(engine)))
		return PTR_ERR(engine);

	/*
	 * Are we already being traced?
	 */
repeat:
	task_lock(current);
	if (!(current->ptrace & PT_PTRACED)) {
		/*
		 * See ptrace_attach() comments about the locking here.
		 */
		unsigned long flags;
		if (!write_trylock_irqsave(&tasklist_lock, flags)) {
			task_unlock(current);
			do {
				cpu_relax();
			} while (!write_can_lock(&tasklist_lock));
			goto repeat;
		}

		ret = security_ptrace_traceme(current->parent);

		/*
		 * Set the ptrace bit in the process ptrace flags.
		 * Then link us on our parent's ptraced list.
		 */
		if (!ret) {
			current->ptrace |= PT_PTRACED;
			__ptrace_link(current, current->real_parent);
		}

		write_unlock_irqrestore(&tasklist_lock, flags);
	}
	task_unlock(current);

	return finish_ptrace_attach(current, engine, ret);
}

/**
 * ptrace_get_task_struct  --  grab a task struct reference for ptrace
 * @pid:       process id to grab a task_struct reference of
 *
 * This function is a helper for ptrace implementations.  It checks
 * permissions and then grabs a task struct for use of the actual
 * ptrace implementation.
 *
 * Returns the task_struct for @pid or an ERR_PTR() on failure.
 */
struct task_struct *ptrace_get_task_struct(pid_t pid)
{
	struct task_struct *child;

	read_lock(&tasklist_lock);
	child = find_task_by_vpid(pid);
	if (child)
		get_task_struct(child);

	read_unlock(&tasklist_lock);
	if (!child)
		return ERR_PTR(-ESRCH);
	return child;
}

#ifndef arch_ptrace_attach
#define arch_ptrace_attach(child)	do { } while (0)
#endif

SYSCALL_DEFINE4(ptrace, long, request, long, pid, long, addr, long, data)
{
	struct task_struct *child;
	long ret;

	/*
	 * This lock_kernel fixes a subtle race with suid exec
	 */
	lock_kernel();
	if (request == PTRACE_TRACEME) {
		ret = ptrace_traceme();
		if (!ret)
			arch_ptrace_attach(current);
		goto out;
	}

	child = ptrace_get_task_struct(pid);
	if (IS_ERR(child)) {
		ret = PTR_ERR(child);
		goto out;
	}

	if (request == PTRACE_ATTACH) {
		ret = ptrace_attach(child);
		/*
		 * Some architectures need to do book-keeping after
		 * a ptrace attach.
		 */
		if (!ret)
			arch_ptrace_attach(child);
		goto out_put_task_struct;
	}

	ret = ptrace_check_attach(child, request == PTRACE_KILL);
	if (ret < 0)
		goto out_put_task_struct;

	ret = arch_ptrace(child, request, addr, data);
	if (ret < 0)
		goto out_put_task_struct;

 out_put_task_struct:
	put_task_struct(child);
 out:
	unlock_kernel();
	return ret;
}

int generic_ptrace_peekdata(struct task_struct *tsk, long addr, long data)
{
	unsigned long tmp;
	int copied;

	copied = access_process_vm(tsk, addr, &tmp, sizeof(tmp), 0);
	if (copied != sizeof(tmp))
		return -EIO;
	return put_user(tmp, (unsigned long __user *)data);
}

int generic_ptrace_pokedata(struct task_struct *tsk, long addr, long data)
{
	int copied;

	copied = access_process_vm(tsk, addr, &data, sizeof(data), 1);
	return (copied == sizeof(data)) ? 0 : -EIO;
}

#if defined CONFIG_COMPAT
#include <linux/compat.h>

int compat_ptrace_request(struct task_struct *child, compat_long_t request,
			  compat_ulong_t addr, compat_ulong_t data)
{
	compat_ulong_t __user *datap = compat_ptr(data);
	compat_ulong_t word;
	siginfo_t siginfo;
	int ret;

	switch (request) {
	case PTRACE_PEEKTEXT:
	case PTRACE_PEEKDATA:
		ret = access_process_vm(child, addr, &word, sizeof(word), 0);
		if (ret != sizeof(word))
			ret = -EIO;
		else
			ret = put_user(word, datap);
		break;

	case PTRACE_POKETEXT:
	case PTRACE_POKEDATA:
		ret = access_process_vm(child, addr, &data, sizeof(data), 1);
		ret = (ret != sizeof(data) ? -EIO : 0);
		break;

	case PTRACE_GETEVENTMSG:
		ret = put_user((compat_ulong_t) child->ptrace_message, datap);
		break;

	case PTRACE_GETSIGINFO:
		ret = ptrace_getsiginfo(child, &siginfo);
		if (!ret)
			ret = copy_siginfo_to_user32(
				(struct compat_siginfo __user *) datap,
				&siginfo);
		break;

	case PTRACE_SETSIGINFO:
		memset(&siginfo, 0, sizeof siginfo);
		if (copy_siginfo_from_user32(
			    &siginfo, (struct compat_siginfo __user *) datap))
			ret = -EFAULT;
		else
			ret = ptrace_setsiginfo(child, &siginfo);
		break;

	default:
		ret = ptrace_request(child, request, addr, data);
	}

	return ret;
}

asmlinkage long compat_sys_ptrace(compat_long_t request, compat_long_t pid,
				  compat_long_t addr, compat_long_t data)
{
	struct task_struct *child;
	long ret;

	/*
	 * This lock_kernel fixes a subtle race with suid exec
	 */
	lock_kernel();
	if (request == PTRACE_TRACEME) {
		ret = ptrace_traceme();
		goto out;
	}

	child = ptrace_get_task_struct(pid);
	if (IS_ERR(child)) {
		ret = PTR_ERR(child);
		goto out;
	}

	if (request == PTRACE_ATTACH) {
		ret = ptrace_attach(child);
		/*
		 * Some architectures need to do book-keeping after
		 * a ptrace attach.
		 */
		if (!ret)
			arch_ptrace_attach(child);
		goto out_put_task_struct;
	}

	ret = ptrace_check_attach(child, request == PTRACE_KILL);
	if (!ret)
		ret = compat_arch_ptrace(child, request, addr, data);

 out_put_task_struct:
	put_task_struct(child);
 out:
	unlock_kernel();
	return ret;
}
#endif	/* CONFIG_COMPAT */
